# -*- coding: utf-8 -*-

import inspect
import itertools
import logging
import os
import sys

# from collections import defaultdict
from collections.abc import Iterable, Sequence
from numbers import Integral
from typing import Literal, NamedTuple, Optional, TypeVar

try:
    from osgeo import ogr, osr

    ogr.UseExceptions()
except ImportError:
    pass
import psycopg2
from shapely import wkb
from shapely.geometry import (
    LinearRing,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from shapely.geometry.base import BaseGeometry

from .geomutils import geom_type_mapping, get_transform_func, unchanged_geom


def setup_logger(
    name: Optional[str] = None, level: int = logging.INFO, show_pid: bool = False
) -> logging.Logger:
    """Setup the logging configuration for a Logger.

    Return a ready-configured logging.Logger instance which will write
    to 'stdout'.


    Keyword arguments:

    name: name of the logging.Logger instance to get. Default is the
    filename where the function is called.
    level: logging level to set to the returned logging.Logger
    instance. Default is logging.INFO.
    show_pid: show the process ID in the log records. Default is
    False.
    """
    if name is None:
        name = os.path.basename(inspect.stack()[1].filename)
    ## Get logger.
    logger = logging.getLogger(name)
    ## Remove existing handlers.
    for handler in logger.handlers:
        logger.removeHandler(handler)
    if level is None:
        logger.disabled = True
        return logger
    ## Set basic logging configuration.
    if show_pid:
        logger.show_pid = True
        pid = f"(PID: {os.getpid()}) "
    else:
        logger.show_pid = False
        pid = ""
    if level <= logging.DEBUG:
        fmt = (
            f"%(asctime)s - %(levelname)s - %(name)s {pid}in %(funcName)s "
            "(l. %(lineno)d) - %(message)s"
        )
    else:
        fmt = "%(asctime)s - %(levelname)s " f"- %(name)s {pid}- %(message)s"
    formatter = logging.Formatter(fmt)
    ch = logging.StreamHandler(sys.stdout)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(level)
    return logger


def update_logger(logger: logging.Logger, **kwargs) -> None:
    """Update the configuration of a logging.Logger instance.


    Positional arguments:

    logger: logging.Logger instance to be updated.

    Keyword arguments:

    Configuration parameters of the logging.Logger instance to update,
    such as "level" or "show_pid". See function "setup_logger".
    """
    level = kwargs.get("level", logger.getEffectiveLevel())
    if level is None:
        logger.disabled = True
        return
    elif "level" in kwargs.keys():
        logger.disabled = False
    ## Set basic logging configuration.
    if not hasattr(logger, "show_pid"):
        logger.show_pid = False
    show_pid = kwargs.get("show_pid", logger.show_pid)
    if show_pid:
        pid = f"(PID: {os.getpid()}) "
    else:
        pid = ""
    if level <= logging.DEBUG:
        fmt = (
            f"%(asctime)s - %(levelname)s - %(name)s {pid}in %(funcName)s "
            "(l. %(lineno)d) - %(message)s"
        )
    else:
        fmt = "%(asctime)s - %(levelname)s " f"- %(name)s {pid}- %(message)s"
    formatter = logging.Formatter(fmt)
    for handler in logger.handlers:
        handler.setFormatter(formatter)
    logger.setLevel(level)


class ConnectionParameters(NamedTuple):
    """Parameters to open a connection to a PostGIS database."""

    host: str
    dbname: str
    user: str
    password: str
    port: int


class SchemaTableColumn(NamedTuple):
    """Location of a geometry column in a PostGIS database."""

    schema: str
    table: str
    column: str


def fetch_geoms_from_pg(
    conn: Optional[psycopg2.extensions.connection] = None,
    conn_params: Optional[ConnectionParameters] = None,
    sql_query: Optional[str] = None,
    geoms_col_loc: Optional[SchemaTableColumn] = None,
    aoi: Optional[BaseGeometry] = None,
    aoi_epsg: Optional[int] = None,
    output_epsg: Optional[int] = None,
):
    """Fetch geometrical features from a PostGIS database.

    Generator function which connects or uses an existing connection
    to a PostGIS database, and yields geometrical features from
    specified geometry column (within a given area or not), or based
    on a user-defined SQL query. If the connection to the database is
    opened by the function, it will be closed automatically after the
    last geometrical feature is yielded.


    Keyword parameters:

    conn: pre-opened connection to the PostGIS database
    ("psycopg2.extensions.connection" instance). Default is None.
    conn_params: connection parameters ("ConnectionParameters"
    instance) to open a connection to the PostGIS database. Default is
    None.
    sql_query: SQL query to use to extract geometrical features from
    the PostGIS database. Note that if this parameter is specified by
    the user, it will override and ignore all other keyword parameters
    except from "conn" and "conn_params". Default is None.
    geoms_col_loc: geometry column location ("SchemaTableColumn"
    instance) within the PostGIS database. Default is None.
    aoi: area of interest, where the geometrical features lies. This
    area must be defined by an instance of the
    "shapely.geometry.base.BaseGeometry" class. Default is None.
    aoi_epsg: EPSG code of the area of interest geometry/ies. Default
    is None.
    output_epsg: EPSG code of the yielded geometrical features. This
    parameter can be used to transform the yielded geometries to a
    different Spatial Reference System from the one used in the
    PostGIS database. Default is None.
    """
    if conn is None:
        if conn_params is None:
            raise ValueError("'conn' and 'conn_params' cannot both be passed None!")
        conn = psycopg2.connect(**conn_params._asdict())
        close_conn = True
    else:
        close_conn = False
    cursor = conn.cursor()
    if sql_query is None:
        if geoms_col_loc is None:
            raise ValueError(
                "'sql_query' and 'geoms_col_loc' cannot both be passed None!"
            )
        if aoi is not None or output_epsg is not None:
            cursor.execute(
                f"SELECT Find_SRID('{geoms_col_loc.schema}', "
                f"'{geoms_col_loc.table}', "
                f"'{geoms_col_loc.column}');"
            )
            pg_epsg = int(cursor.fetchone()[0])
        where_filter = f"WHERE {geoms_col_loc.column} IS NOT NULL"
        if aoi is not None:
            if aoi_epsg is not None and int(aoi_epsg) != pg_epsg:
                transform_aoi = get_transform_func(aoi_epsg, pg_epsg)
                aoi = transform_aoi(aoi)
            spatial_filter = (
                f" AND ST_Intersects({geoms_col_loc.column}, "
                f"ST_GeomFromText('{aoi.wkt}', {pg_epsg}));"
            )
        else:
            spatial_filter = ";"
        if output_epsg is not None and int(output_epsg) != pg_epsg:
            geoms_col_loc = geoms_col_loc._replace(
                column=(f"ST_Transform({geoms_col_loc.column}, " f"{output_epsg})")
            )
        sql_query = (
            f"SELECT ST_AsBinary({geoms_col_loc.column}) "
            f"FROM {geoms_col_loc.schema}.{geoms_col_loc.table} "
            f"{where_filter}{spatial_filter}"
        )
    cursor.execute(sql_query)
    for row in cursor:
        yield wkb.loads(row[0].tobytes())
    if close_conn:
        conn = None


def _get_layer_epsg(layer):
    """Extract and return the EPSG code of an ogr.Layer. Return None
    if not found.
    """
    lyr_srs = layer.GetSpatialRef()
    if lyr_srs is not None and lyr_srs.AutoIdentifyEPSG() == 0:
        return int(lyr_srs.GetAuthorityCode(None))
    else:
        return None


## Type for identifying layers.
LayerID = TypeVar("LayerID", str, int)

## Type for specifing a sequence of layers.
Layers = Sequence[LayerID]


class LayerFilter(NamedTuple):
    """Filter for extraction of geometrical features from file.

    Instances of this class are intended to be used as parameter for
    the "extract_geoms_from_file" function, for filtering and choosing
    the geometrical features to extract.

    Attributes:

    layer_id: LayerID type to identify which layer the filter will be
    applied to. If set to None (default), the filter will be applied
    on all layers.
    aoi: area of interest, where the geometrical features lies. All
    features lying outside the area of interest will be filtered out
    (not extracted). This area must be defined by an instance of the
    "shapely.geometry.base.BaseGeometry" class. Default is None.
    aoi_epsg: EPSG code of the area of interest geometry/ies. Default
    is None (the same Spatial Reference System as the layer will be
    used).
    attr_filter: valid string representation of an attribute filter
    (e.g. "attr_name = 'value'"). Default is None.
    fids: IDs of the features to extract from the layer. This
    parameter will be ignored if either the "aoi" or the "attr_filter"
    parameters are specified by the user.
    """

    layer_id: Optional[LayerID] = None
    aoi: Optional[BaseGeometry] = None
    aoi_epsg: Optional[int] = None
    attr_filter: Optional[str] = None
    fids: Optional[Sequence[int]] = None


## Type for specifying a sequence of LayerFilter instances.
Filters = Sequence[LayerFilter]


def extract_geoms_from_file(
    filename: str,
    driver_name: str,
    layers: Optional[Layers] = None,
    layer_filters: Optional[Filters] = None,
):
    """Extract geometrical features from a GDAL/OGR-readable file.

    Generator function which opens a file located on disk, with one of
    the existing GDAL/OGR drivers, and yields geometrical features,
    from one or several layers. The function also permits the use of
    filters to allow for fine-grained extraction of the geometrical
    features.


    Positional parameters:

    filename: path to the file to extract the geometrical features from.
    driver_name: name of the GDAL/OGR driver to use for opening the
    file. See https://gdal.org/drivers/vector/index.html for a list of
    the supported drivers.

    Keyword parameters:

    layers: Layers type to specify a sequence of layers (identified by
    the "LayerID" type) from which the geometrical features will be
    extracted. Default is None (geometrical features will be extracted
    from all layers).
    layer_filters: Filters type to specify a sequence of filters (see
    "LayerFilter" class) to apply to the layer(s) when extracting the
    geometrical features.
    """
    logger = setup_logger()
    try:
        from osgeo import ogr

        ogr.UseExceptions()
    except ImportError:
        raise NotImplementedError(
            "You must install GDAL/OGR and its Python "
            "bindings to call "
            f"{inspect.stack()[0].function!r}!"
        )
    if not os.path.exists(filename):
        raise ValueError(f"The file {filename!r} does not exist!")
    driver = ogr.GetDriverByName(driver_name)
    if driver is None:
        raise ValueError(
            f"The driver {driver_name!r} is not available or does not exist!"
        )
    ds = driver.Open(filename)
    if layers is not None:
        if not isinstance(layers, Sequence) or isinstance(layers, str):
            raise ValueError(
                "'layers' must be passed an iterable of layer names/indices!"
            )
    else:
        layers = range(ds.GetLayerCount())
    filters_mapping = dict()
    if layer_filters is not None:
        for lf in layer_filters:
            filters_mapping[lf.layer_id] = lf
    for lyr in layers:
        lyr_obj = ds.GetLayer(lyr)
        lyr_filter = filters_mapping.get(lyr, filters_mapping.get(None))
        if lyr_filter is None:
            lyr_aoi = None
            lyr_aoi_epsg = None
            lyr_attr_filter = None
            lyr_fids = None
        else:
            lyr_aoi = lyr_filter.aoi
            lyr_aoi_epsg = lyr_filter.aoi_epsg
            lyr_attr_filter = lyr_filter.attr_filter
            lyr_fids = lyr_filter.fids
        if lyr_aoi is not None:
            if lyr_aoi_epsg is not None:
                lyr_aoi_epsg = int(lyr_aoi_epsg)
                lyr_epsg = _get_layer_epsg(lyr_obj)
                if lyr_epsg is not None and lyr_epsg != lyr_aoi_epsg:
                    transform_aoi = get_transform_func(lyr_aoi_epsg, lyr_epsg)
                    lyr_aoi = transform_aoi(lyr_aoi)
            lyr_obj.SetSpatialFilter(ogr.CreateGeometryFromWkt(lyr_aoi.wkt))
        if lyr_attr_filter is not None:
            lyr_obj.SetAttributeFilter(lyr_attr_filter)
        if lyr_aoi is None and lyr_attr_filter is None and lyr_fids is not None:
            for fid in lyr_fids:
                feature = lyr_obj.GetFeature(fid)
                geom = feature.GetGeometryRef()
                yield wkb.loads(bytes(geom.ExportToWkb()))
        else:
            for feature in lyr_obj:
                geom = feature.GetGeometryRef()
                yield wkb.loads(bytes(geom.ExportToWkb()))
    ds = None


## Type for specifying an Iterable of geometrical features.
GeometryIterable = Iterable[BaseGeometry]


def write_geoms_to_file(
    geoms_iter: GeometryIterable,
    geoms_epsg: int,
    filename: str,
    driver_name: str,
    layer: Optional[LayerID] = None,
    mode: Literal["update", "overwrite"] = "update",
):
    """Write multiple geometrical features to disk.

    The function takes as input an iterable of geometrical features
    and writes them to disk using one of the existing GDAL/OGR
    drivers.


    Positional parameters:

    geoms_iter: iterable of geometrical features
    ("shapely.geometry.base.BaseGeometry" instances).
    geoms_epsg: EPSG code of the input geometrical features. If the
    Spatial Reference System of the input geometrical features is
    different from that of the layer they will written to (in case of
    an update, see "mode" parameter), the coordinates of the
    geometries will be reprojected to the layer's Spatial Reference
    System.
    filename: path to the output file where the geometrical features
    will be written to.

    Keyword parameters:

    layer: layer name on which to write the input geometries. In case
    of a file update (see "mode" parameter), the index of an existing
    layer can be passed as argument. If layer is set to "None"
    (default), the geometrical features will be written, in "update"
    mode, on the first layer available (at index 0), if any. If no
    layer is available, as well as in "overwrite" mode, the layer
    parameter set to "None" will result in the function writing the
    input geometries to a layer named "default" (if the driver
    supports named layers).
    mode: one of "update" or "overwrite". If set to "update", the
    function will update an existing file, or will create it if it
    does not exist. If set to "overwrite", the function will delete
    any file at the path set to the "filename" parameter, and will
    create a new file at this same location.
    """
    logger = setup_logger()
    try:
        from osgeo import ogr, osr

        ogr.UseExceptions()
    except ImportError:
        raise NotImplementedError(
            "You must install GDAL/OGR and its Python bindings to call "
            f"{inspect.stack()[0].function!r}!"
        )
    driver = ogr.GetDriverByName(driver_name)
    geoms_epsg = int(geoms_epsg)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(geoms_epsg)
    geoms_iter = iter(geoms_iter)
    #    geoms_list = iter(geoms_iter)
    #    if not len(set(g.__class__ for g in geoms_list)) == 1:
    #        raise ValueError("Cannot process input geometries of different types!")
    first_geom = next(geoms_iter)
    geom_type = geom_type_mapping[first_geom.geom_type]
    geoms_iter = itertools.chain([first_geom], geoms_iter)
    if not mode in ("update", "overwrite"):
        raise ValueError(
            "Wrong value for the 'mode' argument: must be either 'update' or "
            "'overwrite'!"
        )
    if mode == "update":
        _update_geoms_file(
            geoms_iter, geom_type, geoms_epsg, srs, filename, driver, layer, logger
        )
    else:
        _write_geoms_file(geoms_iter, geom_type, srs, filename, driver, layer, logger)


def _update_geoms_file(
    geoms_iter, geom_type, geoms_epsg, srs, filename, driver, layer, logger
):
    """Update or a create a new file on disk and add input geometries."""
    ds = driver.Open(filename, 1)
    if ds is None:
        _write_geoms_file(
            geoms_iter, geom_type, srs, filename, driver, layer_name, logger
        )
        return
    if layer is not None:
        lyr_obj = ds.GetLayer(layer)
        if lyr_obj is None and isinstance(layer, Integral):
            raise ValueError(f"The layer with index {layer!r} does not exist.")
    else:
        layer = "default"
        lyr_obj = ds.GetLayer()
    transform_geom = unchanged_geom
    if lyr_obj is None:
        lyr_obj = ds.CreateLayer(layer, srs=srs, geom_type=geom_type)
        lyr_def = lyr_obj.GetLayerDefn()
    else:
        lyr_def = lyr_obj.GetLayerDefn()
        lyr_epsg = _get_layer_epsg(lyr_obj)
        if lyr_epsg is not None and lyr_epsg != geoms_epsg:
            logger.info(
                f"The spatial reference system of the output file {filename!r}, "
                f"layer {layer!r}, is different from that of the input geometry "
                "features. The geometry features will be reprojected before being "
                "added to the file."
            )
            transform_geom = get_transform_func(geoms_epsg, lyr_epsg)
        else:
            logger.info(
                f"The spatial reference system of the output file {filename!r}, "
                f"layer {layer!r}, could not be found or identified. Input geometry "
                "features will be added to the file without transformation."
            )
    for geom in geoms_iter:
        feature = ogr.Feature(lyr_def)
        feature.SetGeometry(ogr.CreateGeometryFromWkt(transform_geom(geom).wkt))
        print(lyr_obj.CreateFeature(feature))
        feature = None
    ds = None


def _write_geoms_file(geoms_iter, geom_type, srs, filename, driver, layer, logger):
    """Delete any existing path at given, create a new file on disk
    and add input geometries.
    """
    if isinstance(layer, Integral):
        raise TypeError(
            "You cannot create a new layer by index, you must specify a layer name."
        )
    if os.path.exists(filename):
        driver.DeleteDataSource(filename)
    ds = driver.CreateDataSource(filename)
    if layer is None:
        layer = "default"
    lyr_obj = ds.CreateLayer(layer, srs=srs, geom_type=geom_type)
    lyr_def = lyr_obj.GetLayerDefn()
    for geom in geoms_iter:
        feature = ogr.Feature(lyr_def)
        feature.SetGeometry(ogr.CreateGeometryFromWkt(geom.wkt))
        lyr_obj.CreateFeature(feature)
        feature = None
    ## Close the output file.
    ds = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import inspect
import logging
import multiprocessing as mp
import os
import sqlite3
import time
import uuid
from tempfile import NamedTemporaryFile

import psycopg2
import pyproj
import rtree
import shapely.ops
from pyproj.exceptions import CRSError
from shapely import speedups, wkb

from .comparefunc import geoms_always_match
from .geomrefdb_abc import GeomRefDB
from .geomutils import geom_type_mapping, get_transform_func, to_2D, unchanged_geom
from .io import setup_logger, update_logger
from .misc import SharedIterator, split_iter_to_lists


class PostGISGeomRefDB(GeomRefDB):
    def __init__(self, PG_params, PG_schema, PG_table, PG_geoms_column):

        self.PG_params = PG_params
        self.PG_conn = psycopg2.connect(**PG_params)
        self.PG_schema = PG_schema
        self.PG_table = PG_table
        self.PG_geoms_column = PG_geoms_column

    def __del__(self):
        self.PG_conn = None

    def __getstate__(self):
        attrs = self.__dict__.copy()
        attrs["PG_conn"] = None
        return attrs

    def __setstate__(self, state):
        self.__dict__ = state
        self.PG_conn = psycopg2.connect(**self.PG_params)

    def get_PG_geoms_EPSG(self):
        PG_cursor = self.PG_conn.cursor()
        PG_cursor.execute(
            f"SELECT Find_SRID('{self.PG_schema}', '{self.PG_table}', "
            f"'{self.PG_geoms_column}')"
        )
        res = PG_cursor.fetchone()
        PG_cursor = None
        return int(res[0])

    def true_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        self.logger = setup_logger()
        self.logger.info("Searching true positive geometries...")
        PG_geoms_EPSG = self.get_PG_geoms_EPSG()
        transform = PG_geoms_EPSG != int(geoms_EPSG)
        path2table = ".".join([self.PG_schema, self.PG_table])
        if transform:
            SQL_query_template = (
                f"SELECT ST_AsBinary(ST_Transform({self.PG_geoms_column},"
                f"{geoms_EPSG})) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_Transform(ST_GeomFromText('{{geom.wkt}}', {geoms_EPSG}), "
                f"{PG_geoms_EPSG}))"
            )
        else:
            SQL_query_template = (
                f"SELECT ST_AsBinary({self.PG_geoms_column}) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_GeomFromText('{{geom.wkt}}', {geoms_EPSG}))"
            )
        PG_cursor = self.PG_conn.cursor()
        for geom in geoms_iter:
            PG_cursor.execute(SQL_query_template.format(geom=geom))
            for row in PG_cursor:
                PG_geom = wkb.loads(row[0].tobytes())
                if same_geoms_func(geom, PG_geom):
                    yield geom
                    break
        PG_cursor = None
        logger.info("Done searching true positive geometries.")

    def false_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        logger = setup_logger()
        logger.info("Searching false positive geometries...")
        PG_geoms_EPSG = self.get_PG_geoms_EPSG()
        transform = PG_geoms_EPSG != int(geoms_EPSG)
        path2table = ".".join([self.PG_schema, self.PG_table])
        if transform:
            SQL_query_template = (
                f"SELECT ST_AsBinary(ST_Transform({self.PG_geoms_column},"
                f"{geoms_EPSG})) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_Transform(ST_GeomFromText('{{geom.wkt}}', "
                f"{geoms_EPSG}), {PG_geoms_EPSG}))"
            )
        else:
            SQL_query_template = (
                f"SELECT ST_AsBinary({self.PG_geoms_column}) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_GeomFromText('{{geom.wkt}}', {geoms_EPSG}))"
            )
        PG_cursor = self.PG_conn.cursor()
        for geom in geoms_iter:
            PG_cursor.execute(SQL_query_template.format(geom=geom))
            false_positive = True
            for row in PG_cursor:
                PG_geom = wkb.loads(row[0].tobytes())
                if same_geoms_func(geom, PG_geom):
                    false_positive = False
                    break
            if false_positive:
                yield geom
        PG_cursor = None
        logger.info("Done searching false positive geometries.")

    def missing_geometries(self, geoms_iter, AOI_geom, geoms_EPSG, same_geoms_func):
        logger = setup_logger()
        logger.info("Searching missing geometries...")
        PG_geoms_EPSG = self.get_PG_geoms_EPSG()
        transform = PG_geoms_EPSG != int(geoms_EPSG)
        path2table = ".".join([self.PG_schema, self.PG_table])
        if transform:
            PG_crs = pyproj.CRS(f"EPSG:{PG_geoms_EPSG}")
            AOI_crs = pyproj.CRS(f"EPSG:{geoms_EPSG}")
            project = pyproj.Transformer.from_crs(
                AOI_crs, PG_crs, always_xy=True
            ).transform
            aoi_wkt = shapely.ops.transform(project, AOI_geom).wkt
            SQL_query = (
                f"SELECT ST_AsBinary(ST_Transform({self.PG_geoms_column},"
                f"{geoms_EPSG})) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_GeomFromText('{aoi_wkt}', {PG_geoms_EPSG}))"
            )
        else:
            SQL_query = (
                f"SELECT ST_AsBinary({self.PG_geoms_column}) "
                f"FROM {path2table} "
                f"WHERE ST_Intersects({self.PG_geoms_column}, "
                f"ST_GeomFromText('{AOI_geom.wkt}', {PG_geoms_EPSG}))"
            )
        PG_cursor = self.PG_conn.cursor()
        PG_cursor.execute(SQL_query)
        index = rtree.index.Index()
        for i, geom in enumerate(geoms_iter):
            index.insert(i, geom.bounds, obj=geom)
        for row in PG_cursor:
            PG_geom = wkb.loads(row[0].tobytes())
            if not any(
                same_geoms_func(el.object, PG_geom)
                for el in index.intersection(PG_geom.bounds, objects=True)
            ):
                yield PG_geom
        PG_cursor = None
        logger.info("Done searching missing geometries.")


class RtreeGeomRefDB(GeomRefDB):
    def __init__(self, geoms_iter, geoms_EPSG):
        self.index = rtree.index.Index()
        for i, geom in enumerate(geoms_iter):
            self.index.insert(i, geom.bounds, obj=geom)
        self.EPSG = geoms_EPSG

    def true_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        logger = setup_logger()
        logger.info("Searching true positive geometries...")
        transform = geoms_EPSG != self.EPSG
        if transform:
            geoms_crs = pyproj.CRS(f"EPSG:{geoms_EPSG}")
            ref_DB_crs = pyproj.CRS(f"EPSG:{self.EPSG}")
            project = pyproj.Transformer.from_crs(
                geoms_crs, ref_DB_crs, always_xy=True
            ).transform
            for geom in geoms_iter:
                geom_reproj = shapely.ops.transform(project, geom)
                if any(
                    same_geoms_func(geom_reproj, el.object)
                    for el in self.index.intersection(geom_reproj.bounds, objects=True)
                ):
                    yield geom
        else:
            for geom in geoms_iter:
                if any(
                    same_geoms_func(geom, el.object)
                    for el in self.index.intersection(geom.bounds, objects=True)
                ):
                    yield geom
        logger.info("Done searching true positive geometries.")

    def false_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        logger = setup_logger()
        logger.info("Searching false positive geometries...")
        transform = geoms_EPSG != self.EPSG
        if transform:
            geoms_crs = pyproj.CRS(f"EPSG:{geoms_EPSG}")
            ref_DB_crs = pyproj.CRS(f"EPSG:{self.EPSG}")
            project = pyproj.Transformer.from_crs(
                geoms_crs, ref_DB_crs_crs, always_xy=True
            ).transform
            for geom in geoms_iter:
                geom_reproj = shapely.ops.transform(project, geom)
                if not any(
                    same_geoms_func(geom_reproj, el.object)
                    for el in self.index.intersection(geom_reproj.bounds, objects=True)
                ):
                    yield geom
        else:
            for geom in geoms_iter:
                if not any(
                    same_geoms_func(geom, el.object)
                    for el in self.index.intersection(geom.bounds, objects=True)
                ):
                    yield geom
        logger.info("Done searching false positive geometries.")

    def intersecting_idx_geoms(self, poly=None, bounds=None):
        if poly is not None:
            for el in self.index.intersection(poly.bounds, objects=True):
                idx_geom = el.object
                if poly.intersects(idx_geom):
                    yield idx_geom
        else:
            if bounds is None:
                bounds = self.index.bounds
            for el in self.index.intersection(bounds, objects=True):
                yield el.object

    def missing_geometries(self, geoms_iter, AOI_geom, geoms_EPSG, same_geoms_func):
        logger = setup_logger()
        logger.info("Searching missing geometries...")
        index = rtree.index.Index()
        for i, geom in enumerate(geoms_iter):
            index.insert(i, geom.bounds, obj=geom)
        transform = geoms_EPSG != self.EPSG
        if transform:
            geoms_crs = pyproj.CRS(f"EPSG:{geoms_EPSG}")
            ref_DB_crs = pyproj.CRS(f"EPSG:{self.EPSG}")
            project = pyproj.Transformer.from_crs(
                ref_DB_crs, geoms_crs, always_xy=True
            ).transform
            if AOI_geom is not None:
                project_AOI = pyproj.Transformer.from_crs(
                    geoms_crs, ref_DB_crs, always_xy=True
                ).transform
                AOI_geom = shapely.ops.transform(project_AOI, AOI_geom)
            for ref_geom in self.intersecting_idx_geoms(poly=AOI_geom):
                ref_geom = shapely.ops.transform(project, ref_geom)
                if not any(
                    same_geoms_func(el.object, ref_geom)
                    for el in index.intersection(ref_geom.bounds, objects=True)
                ):
                    yield ref_geom
        else:
            for ref_geom in self.intersecting_idx_geoms(poly=AOI_geom):
                if not any(
                    same_geoms_func(el.object, ref_geom)
                    for el in index.intersection(ref_geom.bounds, objects=True)
                ):
                    yield ref_geom
        logger.info("Done searching missing geometries.")


class SQLiteGeomRefDB(GeomRefDB):
    """Concrete implementation of the GeomRefDB ABC using SQLite.

    SQLiteGeomRefDB is a concrete implementation of the interface
    defined by the GeomRefDB abstract base class. It enables to load
    an existing (or create a new) SQLite database, where geometry
    datasets can be stored and can be compared (based on geometry
    similarity functions) with other geometrical features from an
    external dataset. This class makes use of the spatialite extension
    of SQLite, and as such, spatialite must be installed and available
    in order to work with instances of this class.
    """

    @property
    def supported_geom_types(self):
        return [
            "Point",
            "LineString",
            "Polygon",
            "MultiPoint",
            "MultiLineString",
            "MultiPolygon",
            "GeometryCollection",
        ]

    def __init__(
        self,
        db_path=None,
        default_epsg=None,
        geoms_iter=None,
        geoms_tab_name=None,
        geom_type=None,
        geoms_epsg=None,
        in_ram=True,
        logger=None,
        logger_name=None,
        logging_level=logging.INFO,
    ):
        if db_path is not None:
            self.db_path = os.path.abspath(db_path)
        else:
            self.db_path = None
        if default_epsg is not None:
            try:
                default_epsg = int(default_epsg)
                _ = pyproj.CRS(default_epsg)
            except (CRSError, ValueError, TypeError):
                raise ValueError("{!r} ('default_epsg') is not a valid EPSG code!")
            else:
                self.default_epsg = default_epsg
        else:
            self.default_epsg = None
        self.in_ram = in_ram
        if logger is not None:
            self.logger = logger
        else:
            if logger_name is None:
                logger_name = type(self).__name__
            self.logger = setup_logger(name=logger_name, level=logging_level)

        if db_path is not None and not os.path.isfile(db_path):
            new_db = True
        else:
            new_db = False
        if db_path is None:
            if not in_ram:
                raise ValueError(
                    "The 'db_path' and 'in_ram' parameters cannot both be None!"
                )
            else:
                self.conn = sqlite3.connect(":memory:")
                self.conn.enable_load_extension(True)
                self.conn.load_extension("mod_spatialite")
                cursor = self.conn.cursor()
                cursor.execute("SELECT InitSpatialMetaData();")
                self.conn.commit()
                self.logger.info("New database created in RAM.")
        elif in_ram:
            self.conn = sqlite3.connect(":memory:")
            self.conn.enable_load_extension(True)
            self.conn.load_extension("mod_spatialite")
            if not new_db:
                disk_conn = sqlite3.connect(db_path)
                disk_conn.backup(self.conn)
                disk_conn.close()
                self.logger.info(
                    f"Database file {db_path!r} successfully loaded in RAM."
                )
            else:
                cursor = conn.cursor()
                cursor.execute("SELECT InitSpatialMetaData();")
                self.conn.commit()
                self.logger.info(
                    f"File {dp_path!r} does not exist, new database created in RAM."
                )
        else:
            self.conn = sqlite3.connect(db_path)
            self.conn.enable_load_extension(True)
            self.conn.load_extension("mod_spatialite")
            if new_db:
                cursor = conn.cursor()
                cursor.execute("SELECT InitSpatialMetaData();")
                self.conn.commit()
                self.logger.info(f"New database created at {db_path!r}.")
        if geoms_iter is not None:
            try:
                self.add_geometries(
                    geoms_iter,
                    geom_type=geom_type,
                    geoms_epsg=geoms_epsg,
                    geoms_tab_name=geoms_tab_name,
                )
            except Exception:
                if new_db and not in_ram:
                    self.logger.error(
                        "An error occurred while adding geometries to the database, "
                        f"deleting file {db_path!r}..."
                    )
                    os.remove(db_path)
                else:
                    self.logger.error(
                        "An error occurred while adding geometries to the database..."
                    )
                raise

    def __del__(self):
        self.conn.close()
        if hasattr(self, "db_tf") and os.path.isfile(self.db_tf):
            try:
                os.remove(self.db_tf)
            except PermissionError:
                pass

    def __getstate__(self):
        db_tf = NamedTemporaryFile(suffix=".db", delete=False)
        db_tf.close()
        update_logger(self.logger, level=None)
        self.save_db(db_tf.name)
        update_logger(self.logger, level=self.logger.getEffectiveLevel())
        attrs = self.__dict__.copy()
        attrs["db_tf"] = db_tf.name
        attrs["conn"] = None
        return attrs

    def __setstate__(self, state):
        self.__dict__ = state
        if self.in_ram:
            disk_conn = sqlite3.connect(self.db_tf)
            self.conn = sqlite3.connect(":memory:")
            disk_conn.backup(self.conn)
            disk_conn.close()
        else:
            self.conn = sqlite3.connect(self.db_tf)
        self.conn.enable_load_extension(True)
        self.conn.load_extension("mod_spatialite")

    def save_db(self, path, overwrite=True):
        if os.path.isfile(path):
            if overwrite:
                self.logger.info(f"File {path!r} already exists. Removing file...")
                os.remove(path)
            else:
                self.logger.info(
                    f"File {path!r} already exists. Saving database aborted."
                )
                return
        disk_conn = sqlite3.connect(path)
        self.conn.backup(disk_conn)
        disk_conn.close()
        self.logger.info("Database was saved successfully.")

    def add_geometries(
        self, geoms_iter, geom_type=None, geoms_epsg=None, geoms_tab_name=None
    ):
        self.logger.info("Adding geometries to the database...")
        db_info = self.db_geom_info()
        ## Coordinates of input geometries are not transformed by
        ## default. Return the input geometry unchanged.
        transform_geom = unchanged_geom
        cursor = self.conn.cursor()
        if geom_type is not None and not geom_type in self.supported_geom_types:
            raise ValueError(
                f"{geom_type!r} is not a valid value for the 'geom_type' argument! "
                "Supported geometry types are: {}.".format(
                    ", ".join([f"{gt!r}" for gt in self.supported_geom_types])
                )
            )
        if geoms_epsg is not None:
            try:
                geoms_epsg = int(geoms_epsg)
                _ = pyproj.CRS(geoms_epsg)
            except (CRSError, ValueError, TypeError):
                raise ValueError(f"{geoms_epsg!r} is not a valid EPSG code!")
        if geoms_tab_name is None:
            geoms_tab_name = "default_table"
        tab_info = db_info.get(geoms_tab_name, None)
        if tab_info is None:
            new_tab = True
        else:
            new_tab = False
        if new_tab:
            if geom_type is None:
                raise ValueError(
                    "'geom_type' cannot be passed None for new databases, or if no "
                    "existing geometry table name ('geoms_tab_name') was passed as "
                    "parameter!"
                )
            if geoms_epsg is None:
                if self.default_epsg is not None:
                    geoms_epsg = self.default_epsg
                else:
                    raise ValueError(
                        "'geoms_epsg' cannot be passed None if no default EPSG has "
                        "been set!"
                    )
            cursor.execute(
                f"CREATE TABLE {geoms_tab_name} "
                "(r_id INTEGER PRIMARY KEY AUTOINCREMENT);"
            )
            cursor.execute(
                f"SELECT AddGeometryColumn ('{geoms_tab_name}', "
                f"'geometry', {geoms_epsg}, '{geom_type}', 'XY', 1);"
            )
            cursor.execute(
                f"SELECT CreateSpatialIndex('{geoms_tab_name}', 'geometry');"
            )
            self.conn.commit()

        else:  # if existing table
            if geom_type is not None and geom_type != tab_info["geom_type"]:
                raise ValueError(
                    f"The {geom_type!r} geometry type does not match the geometry type "
                    f"of the {geoms_tab_name!r} table!"
                )
            if geoms_epsg is None:
                geoms_epsg = tab_info["srid"]
            elif geoms_epsg != tab_info["srid"]:
                transform_geom = get_transform_func(geoms_epsg, tab_info["srid"])
                geoms_epsg = tab_info["srid"]
        for geom in geoms_iter:
            cursor.execute(
                f"INSERT INTO {geoms_tab_name} (geometry) VALUES "
                f"(GeomFromText('{transform_geom(to_2D(geom)).wkt}', "
                f"{geoms_epsg}));"
            )
        self.conn.commit()

    def db_geom_info(self, to_stdout=False):

        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM geometry_columns")
        info = {
            tab[0]: {"geom_type": geom_type_mapping[tab[2]], "srid": tab[4]}
            for tab in cursor.fetchall()
        }
        if not to_stdout:
            return info
        elif not info:
            return None
        else:
            f1 = "TABLE"
            f1_width = max(max(len(tab_name) for tab_name in info.keys()), len(f1))
            f2 = "GEOMETRY"
            f2_width = max(max(len(info[k]["geom_type"]) for k in info.keys()), len(f2))
            f3 = "EPSG"
            f3_width = max(max(len(str(info[k]["srid"])) for k in info.keys()), len(f3))
            line_tmp = f"{{:<{f1_width}}} | {{:<{f2_width}}} | {{:>{f3_width}}}\n"
            head = line_tmp.format(f1, f2, f3)
            line_width = len(head) - 1
            table = f"\n{head}{'-' * line_width}\n"
            for k, v in info.items():
                table += line_tmp.format(k, v["geom_type"], str(v["srid"]))
            print(table)

    @staticmethod
    def _get_spatial_query(spatial_index=True, only_within_aoi=False, within_aoi=False):
        query = "SELECT AsBinary(geometry) FROM {{table}}"
        if not spatial_index:
            return query.format() + ";"
        else:
            query += " {spatial_filter};"
        sf = " WHERE {{table}}.ROWID IN {aoi}{geom}"
        if only_within_aoi or within_aoi:
            aoi = (
                "(SELECT ROWID "
                " FROM SpatialIndex "
                " WHERE f_table_name = '{table}' "
                "   AND search_frame = GeomFromText('{aoiwkt}', {epsg})) "
                "AND Intersects(geometry, GeomFromText('{aoiwkt}', {epsg}))"
            )
        else:
            aoi = ""
        if only_within_aoi:
            geom = ""
            return query.format(spatial_filter=sf.format(**locals()))
        else:
            geom = (
                "(SELECT ROWID "
                " FROM SpatialIndex "
                " WHERE f_table_name = '{table}' "
                "   AND search_frame = GeomFromText('{geomsfwkt}', {epsg}))"
            )
        if within_aoi:
            aoi += " AND {table}.ROWID IN "
        return query.format(spatial_filter=sf.format(**locals()))

    def _geoms_generator(
        self,
        geoms_iter,
        transform_geom,
        geoms_match,
        get_search_frame,
        query,
        query_kwargs,
        matching_geoms=True,
    ):
        cursor = self.conn.cursor()
        if matching_geoms:
            for geom in geoms_iter:
                geom_reproj = transform_geom(to_2D(geom))
                search_frame = get_search_frame(geom_reproj)
                query_kwargs["geomsfwkt"] = search_frame.wkt
                cursor.execute(query.format(**query_kwargs))
                if any(geoms_match(geom_reproj, wkb.loads(row[0])) for row in cursor):
                    yield geom
        else:
            for geom in geoms_iter:
                geom_reproj = transform_geom(to_2D(geom))
                search_frame = get_search_frame(geom_reproj)
                query_kwargs["geomsfwkt"] = search_frame.wkt
                cursor.execute(query.format(**query_kwargs))
                if not any(
                    geoms_match(geom_reproj, wkb.loads(row[0])) for row in cursor
                ):
                    yield geom

    def true_positives(
        self,
        geoms_iter,
        aoi_geom=None,
        geoms_epsg=None,
        geoms_tab_name=None,
        geoms_match=None,
        get_search_frame=None,
        ncores=None,
    ):  # , **kwargs):
        self.logger.info("Searching true positive geometries...")
        query_kwargs = dict()
        ## Set default behaviors.
        ## If no function is passed to the 'geoms_match' parameter,
        ## the "test geometry" will always match the "reference
        ## geometry". This way, the user can still use the spatial
        ## index of the internal database and make simple query against
        ## bounding boxes of the stored geometries, without further
        ## assessment of whether geometries match each others.
        if geoms_match is None:
            geoms_match = geoms_always_match
        ## Default search frame is the input geometry itself.
        if get_search_frame is None:
            get_search_frame = unchanged_geom
        ## Coordinates of input geometries are not transformed by
        ## default. Return the input geometry unchanged.
        transform_geom = unchanged_geom
        if geoms_tab_name is None:
            geoms_tab_name = "default_table"
        db_info = self.db_geom_info()
        tab_info = db_info.get(geoms_tab_name, None)
        if tab_info is None:
            raise RuntimeError("No {!r} table was found in the database!")
        else:
            query_kwargs["table"] = geoms_tab_name
        if geoms_epsg is None:
            geoms_epsg = tab_info["srid"]
        else:
            try:
                geoms_epsg = int(geoms_epsg)
                _ = pyproj.CRS(geoms_epsg)
            except (CRSError, ValueError, TypeError):
                raise ValueError(f"{geoms_epsg!r} is not a valid EPSG code!")
            if geoms_epsg != tab_info["srid"]:
                transform_geom = get_transform_func(geoms_epsg, tab_info["srid"])
                if aoi_geom is not None:
                    aoi_geom = transform_geom(aoi_geom)
        query_kwargs["epsg"] = geoms_epsg
        if aoi_geom is not None:
            query_kwargs["aoiwkt"] = aoi_geom.wkt
        if ncores is not None:
            try:
                ncores = int(ncores)
                max_cores = mp.cpu_count() - 1
                if ncores > max_cores:
                    self.logger.info(
                        f"Value {ncores} passed to the 'ncores' argument is too high "
                        "and may result in performance penalty, setting it down to "
                        f"{max_cores}."
                    )
                    ncores = max_cores
                elif ncores < 2:
                    ncores = None
            except (ValueError, TypeError):
                raise ValueError(
                    f"{ncores!r} is not a valid value for the 'ncores' argument!"
                )
        query = self._get_spatial_query(within_aoi=aoi_geom is not None)
        if ncores is not None:
            tp_gen = self._parallelized_method(
                ncores,
                geoms_iter,
                transform_geom,
                geoms_match,
                get_search_frame,
                query,
                query_kwargs,
                matching_geoms=True,
                method_name="_geoms_generator",
            )
        else:
            tp_gen = self._geoms_generator(
                geoms_iter,
                transform_geom,
                geoms_match,
                get_search_frame,
                query,
                query_kwargs,
                matching_geoms=True,
            )
        yield from tp_gen
        self.logger.info("Done searching true positive geometries.")

    def false_positives(
        self,
        geoms_iter,
        aoi_geom=None,
        geoms_epsg=None,
        geoms_tab_name=None,
        geoms_match=None,
        get_search_frame=None,
        ncores=None,
    ):  # , **kwargs):
        self.logger.info("Searching false positive geometries...")
        query_kwargs = dict()
        ## Set default behaviors.
        ## If no function is passed to the 'geoms_match' parameter,
        ## the "test geometry" will always match the "reference
        ## geometry". This way, the user can still use the spatial
        ## index of the internal database and make simple query against
        ## bounding boxes of the stored geometries, without further
        ## assessment of whether geometries match.
        if geoms_match is None:
            geoms_match = geoms_always_match
        ## Default search frame is the input geometry itself.
        if get_search_frame is None:
            get_search_frame = unchanged_geom
        ## Coordinates of input geometries are not transformed by
        ## default. Return the input geometry unchanged.
        transform_geom = unchanged_geom
        if geoms_tab_name is None:
            geoms_tab_name = "default_table"
        db_info = self.db_geom_info()
        tab_info = db_info.get(geoms_tab_name, None)
        if tab_info is None:
            raise RuntimeError("No {!r} table was found in the database!")
        else:
            query_kwargs["table"] = geoms_tab_name
        if geoms_epsg is None:
            geoms_epsg = tab_info["srid"]
        else:
            try:
                geoms_epsg = int(geoms_epsg)
                _ = pyproj.CRS(geoms_epsg)
            except (CRSError, ValueError):
                raise ValueError(f"{geoms_epsg!r} is not a valid EPSG code!")
            if geoms_epsg != tab_info["srid"]:
                transform_geom = get_transform_func(geoms_epsg, tab_info["srid"])
                if aoi_geom is not None:
                    aoi_geom = transform_geom(aoi_geom)
        query_kwargs["epsg"] = geoms_epsg
        if aoi_geom is not None:
            query_kwargs["aoiwkt"] = aoi_geom.wkt
        if ncores is not None:
            try:
                ncores = int(ncores)
                max_cores = mp.cpu_count() - 1
                if ncores > max_cores:
                    self.logger.info(
                        f"Value {ncores} passed to the 'ncores' argument is too high "
                        "and may result in performance penalty, setting it down to "
                        f"{max_cores}."
                    )
                    ncores = max_cores
                elif ncores < 2:
                    ncores = None
            except (ValueError, TypeError):
                raise ValueError(
                    f"{ncores!r} is not a valid value for the 'ncores' argument!"
                )
        query = self._get_spatial_query(within_aoi=aoi_geom is not None)
        if ncores is not None:
            fp_gen = self._parallelized_method(
                ncores,
                geoms_iter,
                transform_geom,
                geoms_match,
                get_search_frame,
                query,
                query_kwargs,
                matching_geoms=False,
                method_name="_geoms_generator",
            )
        else:
            fp_gen = self._geoms_generator(
                geoms_iter,
                transform_geom,
                geoms_match,
                get_search_frame,
                query,
                query_kwargs,
                matching_geoms=False,
            )
        yield from fp_gen
        self.logger.info("Done searching false positive geometries.")

    def missing_geometries(
        self,
        geoms_iter,
        geom_type=None,
        aoi_geom=None,
        geoms_epsg=None,
        geoms_tab_name=None,
        geoms_match=None,
        get_search_frame=None,
        ncores=None,
    ):  # , **kwargs):
        self.logger.info("Searching missing geometries...")
        query_kwargs = dict()
        if geoms_tab_name is None:
            geoms_tab_name = "default_table"
        db_info = self.db_geom_info()
        tab_info = db_info.get(geoms_tab_name, None)
        if tab_info is None:
            raise RuntimeError("No {!r} table was found in the database!")
        if geom_type is None:
            self.logger.info(
                "No geometry type was passed to the 'geom_type' argument. Assuming "
                "that all input geometries have the same geometry type as the ones "
                f"stored in the {geoms_tab_name!r} table."
            )
            geom_type = tab_info["geom_type"]
        query_kwargs["table"] = geoms_tab_name
        db_epsg = tab_info["srid"]
        query_kwargs["epsg"] = db_epsg
        if geoms_epsg is None:
            geoms_epsg = db_epsg
        else:
            try:
                geoms_epsg = int(geoms_epsg)
                _ = pyproj.CRS(geoms_epsg)
            except (CRSError, ValueError, TypeError):
                raise ValueError(f"{geoms_epsg!r} is not a valid EPSG code!")
            if geoms_epsg != db_epsg and aoi_geom is not None:
                transform_aoi = get_transform_func(geoms_epsg, db_epsg)
                aoi_geom = transform_aoi(aoi_geom)
        if aoi_geom is not None:
            query_kwargs["aoiwkt"] = aoi_geom.wkt
        logger_conf = dict()
        if ncores is not None:
            try:
                ncores = int(ncores)
                max_cores = mp.cpu_count() - 1
                if ncores > max_cores:
                    self.logger.info(
                        f"Value {ncores} passed to the 'ncores' argument is too high "
                        "and may result in performance penalty, setting it down to "
                        f"{max_cores}."
                    )
                    ncores = max_cores
                elif ncores < 2:
                    ncores = None
            except (ValueError, TypeError):
                raise ValueError(
                    f"{ncores!r} is not a valid value for the 'ncores' argument!"
                )
        if ncores is None:
            logger_conf.update(
                {"logger_name": str(uuid.uuid1()), "logging_level": None}
            )
        else:
            logger_conf["logger"] = self.logger

        input_geoms_db = SQLiteGeomRefDB(
            geoms_iter=geoms_iter,
            geom_type=geom_type,
            geoms_epsg=geoms_epsg,
            **logger_conf,
        )
        query = self._get_spatial_query(only_within_aoi=aoi_geom is not None)
        cursor = self.conn.cursor()
        cursor.execute(query.format(**query_kwargs))
        geoms_iter = (wkb.loads(row[0]) for row in cursor)
        if ncores is not None:
            mg_gen = input_geoms_db._parallelized_method(
                ncores,
                geoms_iter,
                method_name="false_positives",
                geoms_epsg=db_epsg,
                geoms_match=geoms_match,
                get_search_frame=get_search_frame,
            )
        else:

            mg_gen = input_geoms_db.false_positives(
                geoms_iter,
                geoms_epsg=db_epsg,
                geoms_match=geoms_match,
                get_search_frame=get_search_frame,
            )
        yield from mg_gen
        del input_geoms_db
        self.logger.info("Done searching missing geometries.")

    def _parallelized_method(
        self, ncores, geoms_iter, *method_args, method_name="", **method_kwargs
    ):
        if not method_name:
            method_name = inspect.stack()[1][3]
        method_obj = getattr(self, method_name, None)
        if method_obj is None:
            raise ValueError(
                f"{method_name!r} is not a valid method name for "
                f"the class {cls.__name__!r}!"
            )
        logger_conf = {
            "name": self.logger.name,
            "level": self.logger.getEffectiveLevel(),
        }
        method_kwargs["logger_conf"] = logger_conf
        geoms_iters = split_iter_to_lists(geoms_iter, ncores)
        shared_results_iter = SharedIterator()
        nprocs_done = mp.Value("i", 0)
        procs = list()
        for i in range(ncores):
            procs.append(
                mp.Process(
                    target=self._wrap_method_return,
                    args=[
                        method_obj,
                        geoms_iters[i],
                        shared_results_iter,
                        nprocs_done,
                        *method_args,
                    ],
                    kwargs=method_kwargs,
                )
            )
        try:
            for p in procs:
                p.start()
                self.logger.info(f"New process spawned (PID: {p.pid}).")
            ## Must check sentinel value, join cannot be called as the
            ## spawned processes use a queue object (see:
            ## https://docs.python.org/3/library/multiprocessing.html#programming-guidelines).
            while not nprocs_done.value == ncores:
                time.sleep(0.2)
        except KeyboardInterrupt:
            for p in procs:
                p.terminate()
                p.join()
            del geoms_iters
            raise
        else:
            return shared_results_iter

    def _wrap_method_return(
        self,
        method_obj,
        geoms_iter,
        shared_results_iter,
        nprocs_done,
        *method_args,
        logger_conf=None,
        **method_kwargs,
    ):
        if logger_conf is None:
            logger = setup_logger(name=str(uuid.uuid1()), level=None)
        else:
            logger_conf["name"] = (
                logger_conf.get("name", type(self).__name__) + f" (PID: {os.getpid()})"
            )
            logger = setup_logger(**logger_conf)
        logger.info("Start worker function.")
        res = list(method_obj(geoms_iter, *method_args, **method_kwargs))
        n_geoms = len(res)
        if n_geoms > 1:
            logger.info(f"{n_geoms} geometries found.")
        else:
            logger.info(f"{n_geoms} geometry found.")
        shared_results_iter.put_iter(res)
        with nprocs_done.get_lock():
            nprocs_done.value += 1
        logger.info("Worker function completed.")



# -*- coding: utf-8 -*-



from functools import partial

import numpy as np


dispatch_funcs = list()

def dispatch_function(func):
    dispatch_funcs.append(func.__name__)
    return func

def geoms_always_match(gtest, gref):
    return True

def _perc_area_ptest_overlap(gtest, gref, threshold=None):
    intersection = gtest.intersection(gref)
    return (intersection.area / gtest.area) >= threshold

def _perc_area_pref_overlap(gtest, gref, threshold=None):
    intersection = gtest.intersection(gref)
    return (intersection.area / gref.area) >= threshold

def _perc_area_both_overlap(gtest, gref, threshold=None):
    intersection = gtest.intersection(gref)
    return min(intersection.area / gtest.area,
               intersection.area / gref.area) >= threshold

def _mean_perc_area_overlap(gtest, gref, threshold=None):
    intersection = gtest.intersection(gref)
    return np.mean([intersection.area / gtest.area,
                    intersection.area / gref.area]) >= threshold

def _intersection_over_union(gtest, gref, threshold=None):
    intersection = gtest.intersection(gref)
    union = gtest.union(gref)
    return (intersection.area / union.area) >= threshold


_strategy_mapping = {"ptest": _perc_area_ptest_overlap,
                     "pref": _perc_area_pref_overlap,
                     "both": _perc_area_both_overlap,
                     "mean": _mean_perc_area_overlap,
                     "IoU": _intersection_over_union}

@dispatch_function
def polygons_area_match(strategy, threshold):
    if strategy not in _strategy_mapping.keys():
        raise ValueError("The strategy parameter must be passed one of {}!"
                         .format(", ".join([f"{s!r}"
                                            for s in _strategy_mapping.keys()])))
    try:
        assert 0 < float(threshold) <= 1
    except (AssertionError, ValueError):
        raise ValueError("The threshold parameter must passed a floating point "
                         "number between 0.0 (excluded) and 1.0 (included)!")
    return partial(_strategy_mapping[strategy], threshold=float(threshold))




# -*- coding: utf-8 -*-


def recall_score(tp_num: int, mg_num: int):
    return tp_num / (tp_num + mg_num)


def precision_score(tp_num: int, fp_num: int):
    return tp_num / (tp_num + fp_num)


def f1_score(tp_num: int, fp_num: int, mg_num: int):
    return tp_num / (tp_num + ((mg_num + fp_num) / 2))



# -*- coding: utf-8 -*-


import queue
import multiprocessing as mp


def gen_from_queue(q, lock):
    try:
        with lock:
            yield q.get(block=False)
    except queue.Empty:
        raise StopIteration

def iter_to_queue(it, q, lock):
    with lock:
        for el in iter(it):
            q.put((el))


class SharedIterator:

    @property
    def lock(self):
        return self._l

    @property
    def size(self):
        with self._s.get_lock():
            size = self._s.value
        return size

    def __init__(self, it=None, lock=None):
        self._q = mp.Queue()
        self._l = mp.Lock() if lock is None else lock
        self._s = mp.Value("i", 0)
        if it is not None:
            for count, el in enumerate(it, 1):
                self._q.put(el)
            self._s.value += count


    def __next__(self):

        while self.size:
            try:
                el = self._q.get(block=False)
            except queue.Empty:
                continue
            with self._s.get_lock():
                self._s.value -= 1
            return el
        raise StopIteration()

    
    def __iter__(self):
        return self


    def put(self, el):
        self._q.put(el)
        with self._s.get_lock():
            self._s.value += 1

    def put_iter(self, it):
        with self._l:
            for count, el in enumerate(it, 1):
                self._q.put(el)
        with self._s.get_lock():
            self._s.value += count           


def split_iter_to_lists(it, nlists):
    lists = [[] for _ in range(nlists)]
    for i,el in enumerate(it):
        lists[i % nlists].append(el)
    return lists
        



# -*- coding: utf-8 -*-

from shapely.geometry import Point, Polygon
import pytest

from geomcompare.geomutils import to_2D, unchanged_geom


@pytest.fixture
def geoms_3D():
    pt3D = Point((0.0, 1, 4.5))
    poly3D = Polygon(((0.0, 0.0, 0.0), (1.2, 0.5, -1), (1.5, 1, -2.1)))
    return [pt3D, poly3D]


def test_to_2D(geoms_3D):
    assert all(g.has_z for g in geoms_3D)
    assert all(not g.has_z for g in map(to_2D, geoms_3D))


def test_unchanged_geom(geoms_3D):
    assert all(g.equals(unchanged_geom(g)) for g in geoms_3D)



# -*- coding: utf-8 -*-


import abc
from collections.abc import Sequence
import multiprocessing as mp
import signal

from .stats import recall_score, precision_score, f1_score


class GeomRefDB(abc.ABC):


    @abc.abstractmethod
    def true_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        """ Return an iterable of input geometries that are matching
        geometries of the GeomRefDB instance.
        """
            

    @abc.abstractmethod
    def false_positives(self, geoms_iter, geoms_EPSG, same_geoms_func):
        """ Return an iterable of input geometries that are not matching
        any geometry of the GeomRefDB instance.
        """


    @abc.abstractmethod
    def missing_geometries(self, geoms_iter, AOI_geom, geoms_EPSG,
                           same_geoms_func):
        """ Return an iterable of geometries of the GeomRefDB that are
        not matching any of the input geometries.
        """


    def get_recall_score(self, geoms_iter, AOI_geom, geoms_EPSG,
                         same_geoms_func):
        if not isinstance(geoms_iter, Sequence):
            geoms_iter = list(geoms_iter)
        try:
            tp_num = len(self.true_positives(geoms_iter,
                                             geoms_epsg=geoms_EPSG,
                                             geoms_match=same_geoms_func))
        except TypeError:
            tp_num = sum(1 for _ in self.true_positives(geoms_iter,
                                                        geoms_epsg=geoms_EPSG,
                                                        geoms_match=same_geoms_func))
        try:
            mg_num = len(self.missing_geometries(geoms_iter,
                                                 aoi_geom=AOI_geom,
                                                 geoms_epsg=geoms_EPSG,
                                                 geoms_match=same_geoms_func))
        except TypeError:
            mg_num = sum(1 for _ in self.missing_geometries(geoms_iter,
                                                            aoi_geom=AOI_geom,
                                                            geoms_epsg=geoms_EPSG,
                                                            geoms_match=same_geoms_func))
        return recall_score(tp_num, mg_num)

        
    def get_precision_score(self, geoms_iter, geoms_EPSG, same_geoms_func):
        if not isinstance(geoms_iter, Sequence):
            geoms_iter = list(geoms_iter)
        try:
            tp_num = len(self.true_positives(geoms_iter,
                                             geoms_epsg=geoms_EPSG,
                                             geoms_match=same_geoms_func))
        except TypeError:
            tp_num = sum(1 for _ in self.true_positives(geoms_iter,
                                                        geoms_epsg=geoms_EPSG,
                                                        geoms_match=same_geoms_func))
        try:
            fp_num = len(self.false_positives(geoms_iter,
                                              geoms_epsg=geoms_EPSG,
                                              geoms_match=same_geoms_func))
        except TypeError:
            fp_num = sum(1 for _ in self.false_positives(geoms_iter,
                                                         geoms_epsg=geoms_EPSG,
                                                         geoms_match=same_geoms_func))            
        return precision_score(tp_num, fp_num)


    def get_f1_score(self, geoms_iter, AOI_geom, geoms_EPSG,
                     same_geoms_func):
        if not isinstance(geoms_iter, Sequence):
            geoms_iter = list(geoms_iter)
        try:
            tp_num = len(self.true_positives(geoms_iter,
                                             geoms_epsg=geoms_EPSG,
                                             geoms_match=same_geoms_func))
        except TypeError:
            tp_num = sum(1 for _ in self.true_positives(geoms_iter,
                                                        geoms_epsg=geoms_EPSG,
                                                        geoms_match=same_geoms_func))
        try:
            fp_num = len(self.false_positives(geoms_iter,
                                              geoms_epsg=geoms_EPSG,
                                              geoms_match=same_geoms_func))
        except TypeError:
            fp_num = sum(1 for _ in self.false_positives(geoms_iter,
                                                         geoms_epsg=geoms_EPSG,
                                                         geoms_match=same_geoms_func))
        try:
            mg_num = len(self.missing_geometries(geoms_iter,
                                                 aoi_geom=AOI_geom,
                                                 geoms_epsg=geoms_EPSG,
                                                 geoms_match=same_geoms_func))
        except TypeError:
            mg_num = sum(1 for _ in self.missing_geometries(geoms_iter,
                                                            aoi_geom=AOI_geom,
                                                            geoms_epsg=geoms_EPSG,
                                                            geoms_match=same_geoms_func))
            return f1_score(tp_num, fp_num, mg_num)


    def compare_full(self, geoms_iter, AOI_geom, geoms_EPSG, same_geoms_func):
        if not isinstance(geoms_iter, Sequence):
            geoms_iter = list(geoms_iter)
        results = dict()
        results['true_positives'] = list(self.true_positives(geoms_iter,
                                                            geoms_epsg=geoms_EPSG,
                                                            geoms_match=same_geoms_func))
        results['false_positives'] = list(self.false_positives(geoms_iter,
                                                               geoms_epsg=geoms_EPSG,
                                                               geoms_match=same_geoms_func))
        results['missing_geometries'] = list(self.missing_geometries(geoms_iter,
                                                                      aoi_geom=AOI_geom,
                                                                      geoms_epsg=geoms_EPSG,
                                                                      geoms_match=same_geoms_func))
        tp_num = len(results['true_positives'])
        fp_num = len(results['false_positives'])
        mg_num = len(results['missing_geometries'])
        results['recall'] = recall_score(tp_num, mg_num)
        results['precision'] = precision_score(tp_num, fp_num)
        results['f1'] = f1_score(tp_num, fp_num, mg_num)
        return results
            



# -*- coding: utf-8 -*-


from functools import partial
import shapely.ops
import pyproj

geom_type_mapping = {"LinearRing": 101, # ogr.wkbLinearRing
                     "LineString": 2, # ogr.wkbLineString
                     "MultiLineString": 5, #ogr.wkbMultiLineString
                     "MultiPoint": 4, # ogr.wkbMultiPoint
                     "MultiPolygon": 6, # ogr.wkbMultiPolygon
                     "Point": 1, # ogr.wkbPoint
                     "Polygon": 3, # ogr.wkbPolygon
                     "GeometryCollection": 7, # ogr.wkbGeometryCollection
                     }
geom_type_mapping.update({v: k for k,v in geom_type_mapping.items()})


to_2D = partial(shapely.ops.transform, lambda *geom_coords: geom_coords[:2])

def get_transform_func(epsg_in, epsg_out):
    crs_in = pyproj.CRS("EPSG:{}".format(epsg_in))
    crs_out = pyproj.CRS("EPSG:{}".format(epsg_out))
    project = pyproj.Transformer.from_crs(crs_in, crs_out,
                                          always_xy=True).transform
    return partial(shapely.ops.transform, project)

def unchanged_geom(geom):
    return geom
